/**************************
 * (C) 2013 Kelley Hecker *
 *  All rights reserved   *
 **************************/

// Contains dimensions of the grid and stores targets. Also 
// converts grid points to openGL points and assigns colors. 
// Checks collisions.

#include "GameBoard.h"
#include "Parameters.h"

using namespace std;

// Construct a game board with specific pixel width and height, grid width and height, segment 
// and target colors.
GameBoard::GameBoard(int pWidth, int pHeight, int gWidth, int gHeight, Color _segmentColors[], int numSegmentColors, Color _targetColors[], int numTargetColors) {
	pixelWidth = pWidth;
	pixelHeight = pHeight;
	gridWidth = gWidth;
	gridHeight = gHeight;
	segmentColors = vector<Color>(_segmentColors, _segmentColors + numSegmentColors);
	targetColors = vector<Color>(_targetColors, _targetColors + numTargetColors);

	cellWidth = 1.0f/gridWidth;
	cellHeight = 1.0f/gridHeight;

	AddTarget();
}

// Get the color of a segment with given color index.
Color GameBoard::GetSegmentColorAtIndex(int index) {
	return segmentColors[index];
}

// Get the color of a target with given color index. 
Color GameBoard::GetTargetColorAtIndex(int index) {
	return targetColors[index];
}

// Add a target at a random point on the grid.
void GameBoard::AddTarget() {
	targets.push_back(GridPoint(rand() % gridWidth, rand() % gridHeight, rand() % targetColors.size()));
}

// Convert segments from grid points to openGL points.
vector<Sprite> GameBoard::ConvertGridPointsToSprites(vector<GridPoint> gridPoints) {
	vector<Sprite> spriteVector;

	for (unsigned int n = 0; n < gridPoints.size(); n++)
	{
		float x = 1.0f*gridPoints[n].x/gridWidth + cellWidth/2.0f;
		float y = 1.0f*gridPoints[n].y/gridHeight + cellHeight/2.0f;

		spriteVector.push_back(Sprite(x, y, cellWidth * SEGMENT_SCALE, cellHeight * SEGMENT_SCALE, gridPoints[n].colorIndex));
	}

	return spriteVector;
}

// Convert targets from grid points to openGL points.
vector<Sprite> GameBoard::ConvertTargetsToSprites() {
	vector<Sprite> targetSprites;

	for (unsigned int n = 0; n < targets.size(); n++)
	{
		float x = 1.0f*targets[n].x/gridWidth + cellWidth/2.0f;
		float y = 1.0f*targets[n].y/gridHeight + cellHeight/2.0f;

		targetSprites.push_back(Sprite(x, y, cellWidth * TARGET_SCALE, cellHeight * TARGET_SCALE, targets[n].colorIndex));
	}

	return targetSprites;
}

// Check if there is a target at the given point.
bool GameBoard::CheckHitTarget(int x, int y) {
	for (unsigned int n = 0; n < targets.size(); n++)
	{
		if (targets[n].x == x && targets[n].y == y)
		{
			targets.erase(targets.begin() + n);
			return true;
		}
	}

	return false;
}

// Get if there is a wall at the given point.
bool GameBoard::CheckHitWall(int x, int y) {
	if (x < 0 || x > gridWidth)
		return true;
	if (y < 0 || y > gridHeight)
		return true;

	return false;
}

// Check if the head intersects with any other segments.
bool GameBoard::CheckHitSelf(vector<GridPoint> segments)
{
	int headX = segments[0].x;
	int headY = segments[0].y;

	for (unsigned int n = 1; n < segments.size(); n++)
	{
		if (segments[n].x == headX && segments[n].y == headY)
			return true;
	}

	return false;
}

// Reset the game board.
void GameBoard::Reset() {
	targets.clear();
	AddTarget();
}