/**************************
 * (C) 2013 Kelley Hecker *
 *  All rights reserved   *
 **************************/

// Entry point of the application. Handles all game logic and openGL/GLUT calls.

#define _USE_MATH_DEFINES

#include <stdlib.h>
#include <glut.h>
#include <windows.h>
#include <math.h>
#include <time.h>
#include "GameBoard.h"
#include "Caterpillar.h"
#include "GameController.h"
#include "Parameters.h"

using namespace std;

GameBoard board(PIXEL_WIDTH, PIXEL_HEIGHT, GRID_WIDTH, GRID_HEIGHT, SEGMENT_COLORS, NUM_SEGMENT_COLORS, TARGET_COLORS, NUM_TARGET_COLORS);
Caterpillar caterpillar(GRID_WIDTH/2, GRID_HEIGHT/4, NUM_SEGMENT_COLORS);

int score = 0;

int windowID;

DWORD currentTime;
State currentState;

vector<Sprite> sprites;

// Initialize GLUT view.
void InitializeView(int width, int height) {
	glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA);
	glutInitWindowSize(width, height);
	glutInitWindowPosition(200, 200);
	windowID = glutCreateWindow("Caterpillar");
}

// Initialize openGL.
void InitializeOpenGL() {
	glClearColor(BACKGROUND_COLOR.r, BACKGROUND_COLOR.g, BACKGROUND_COLOR.b, BACKGROUND_COLOR.a);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);

	// Smooth lines.
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glEnable(GL_LINE_SMOOTH);
}

// Logic loop for the main menu.
void MainMenuLoop() {
	currentTime = GetTickCount();
	int millisecondsPerFrame = 1000 * 0.5f;
	float extraTime;

	// Draw the main menu.
	DrawMainMenu((currentTime / millisecondsPerFrame) % 2);

	currentTime += millisecondsPerFrame;
	extraTime = currentTime - GetTickCount();

	// Sleep if we have extra time in this frame.
	if (extraTime > 0)
		Sleep (extraTime);
}

// Logic loop for the game.
void GameMainLoop() {
	currentTime = GetTickCount();
	int millisecondsPerFrame = 1000/FPS;
	float extraTime;

	GridPoint head = caterpillar.GetSegment(0);

	// Check for collisions.
	if (board.CheckHitTarget(head.x, head.y))
	{
		score++;
		board.AddTarget();
		caterpillar.AddSegment();
	}
	else if (board.CheckHitWall(head.x, head.y) || board.CheckHitSelf(caterpillar.GetSegments()))
	{
		GameOver();
		return;
	}

	// Update the segments.
	caterpillar.MoveSegments();

	// Convert grid to sprites for openGL.
	sprites = board.ConvertGridPointsToSprites(caterpillar.GetSegments());

	// Draw the game.
	DrawGame();

	currentTime += millisecondsPerFrame;
	extraTime = currentTime - GetTickCount();

	// Sleep if we have extra time in this frame.
	if (extraTime > 0)
		Sleep (extraTime);
}

// Switches to game over state.
void GameOver() {
	currentState = GAME_OVER;

	// Stop GLUT callbacks.
	glutIdleFunc(NULL);
	glutSpecialFunc(NULL);

	// Draw game over message.
	DrawGameOver();
}

// Resets game and changes to game state.
void ResetGame() {
	// Switch to game state.
	currentState = GAME;
	glutIdleFunc(GameMainLoop);
	glutSpecialFunc(CheckArrowInput);

	// Reset everything.
	board.Reset();
	caterpillar.Reset(GRID_WIDTH/2, GRID_HEIGHT/4);
	score = 0;
}

// Draws the main menu. Letter position is used for title animation.
void DrawMainMenu(int letterPosition) {
	float scale;
	float lineWidth;
	unsigned int wordBeginIndex = 0;

	glClear(GL_COLOR_BUFFER_BIT);

	// Draw the title.
	glPushMatrix();

	glTranslatef(0.0f, 0.75f, 0.0f);
	scale = 0.001f;
	glScalef(scale, scale, scale);
	lineWidth = GetLineWidth(TITLE, 0, TITLE.length());
	glTranslatef((0.5f/scale) - (lineWidth/2.0f), 0.0f, 0.0f);
	glLineWidth(5.0f);

	for (unsigned int n = 0; n < TITLE.length(); n++)
	{
		if ((n + letterPosition)%2 == 0)
			glTranslatef(0.0f, -0.02f / scale, 0.0f);
		else
			glTranslatef(0.0f, 0.02f / scale, 0.0f);

		Color fontColor = board.GetSegmentColorAtIndex(n % NUM_SEGMENT_COLORS);
		glColor3f(fontColor.r, fontColor.g, fontColor.b);
		glutStrokeCharacter(GLUT_STROKE_ROMAN, TITLE[n]);

		if ((n + letterPosition)%2 == 0)
			glTranslatef(0.0f, 0.02f / scale, 0.0f);
		else
			glTranslatef(0.0f, -0.02f / scale, 0.0f);
	}

	glPopMatrix();

	// Draw start instuctions.
	glColor3f(TEXT_COLOR.r, TEXT_COLOR.g, TEXT_COLOR.b);
	glPushMatrix();

	glTranslatef(0.0f, 0.5f, 0.0f);
	scale = 0.0005f;
	glScalef(scale, scale, scale);
	lineWidth = GetLineWidth(START, 0, START.length());
	glTranslatef((0.5f/scale) - (lineWidth/2.0f), 0.0f, 0.0f);
	glLineWidth(4.0f);

	for (unsigned int n = 0; n < START.length(); n++)
		glutStrokeCharacter(GLUT_STROKE_ROMAN, START[n]);

	glPopMatrix();

	// Draw game instructions.
	glPushMatrix();

	glTranslatef(0.0f, 0.4f, 0.0f);
	scale = 0.00035f;
	glScalef(scale, scale, scale);
	glLineWidth(3.0f);

	for (unsigned int n = 0; n < INSTRUCTIONS.length();)
	{
		unsigned int lineEnd = INSTRUCTIONS.find('\n', n);

		if (lineEnd == -1)
			lineEnd = INSTRUCTIONS.length();
		
		lineWidth = GetLineWidth(INSTRUCTIONS, n, lineEnd);

		glTranslatef((0.5f/scale) - (lineWidth/2.0f), -0.075f/scale, 0.0f);

		for (unsigned int m = wordBeginIndex; m < lineEnd; m++)
			glutStrokeCharacter(GLUT_STROKE_ROMAN, INSTRUCTIONS[m]);

		glTranslatef(-lineWidth, 0.0f, 0.0f);
		glTranslatef(-((0.5f/scale) - (lineWidth/2.0f)), 0.0f, 0.0f);

		wordBeginIndex = lineEnd+1;
		n = lineEnd + 1;
	}

	glPopMatrix();

	glFlush();
}

// Draws the game.
void DrawGame() {
	glClear(GL_COLOR_BUFFER_BIT);

	// Draw all the segments.
	for (unsigned int n = 0; n < sprites.size(); n++) {
		
		Color segmentColor = board.GetSegmentColorAtIndex(sprites[n].colorIndex);
		float radiusX = sprites[n].pixelWidth/2.0f;
		float radiusY = sprites[n].pixelHeight/2.0f;

		// Draw the head with antenna.
		if (n == 0)
		{
			glPushMatrix();

			glTranslatef( sprites[n].pixelX, sprites[n].pixelY, 0.0f);

			Direction dir = caterpillar.GetDirection();
			switch (dir) 
			{
			case UP:
				glRotatef(0.0f, 0.0f, 0.0f, 1.0f);
				break;
			case DOWN:
				glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
				break;
			case LEFT:
				glRotatef(90.0f, 0.0f, 0.0f, 1.0f);
				break;
			case RIGHT:
				glRotatef(270.0f, 0.0f, 0.0f, 1.0f);
				break;
			}

			glTranslatef( -sprites[n].pixelX, -sprites[n].pixelY, 0.0f);

			glLineWidth(3.0f);
			glColor3f(ANTENNA_COLOR.r, ANTENNA_COLOR.g, ANTENNA_COLOR.b);

			// Draw right antenna.
			glBegin(GL_LINES);
				glVertex3f(sprites[n].pixelX, sprites[n].pixelY, 0.0f);
				glVertex3f(sprites[n].pixelX + radiusX, sprites[n].pixelY + radiusY, 0.0f); 
			glEnd();

			// Draw left antenna.
			glBegin(GL_LINES);
				glVertex3f(sprites[n].pixelX, sprites[n].pixelY, 0.0f);
				glVertex3f(sprites[n].pixelX - radiusX, sprites[n].pixelY + radiusY, 0.0f); 
			glEnd();

			glPopMatrix();
		}

		glColor3f(segmentColor.r, segmentColor.g, segmentColor.b);

		// Draw circles using polygons.
		glBegin(GL_POLYGON);
			for (double i = 0; i < 2 * M_PI; i += M_PI / CIRCLE_SEGMENTS)
				glVertex3f(sprites[n].pixelX + cos(i) * radiusX, sprites[n].pixelY + sin(i) * radiusY, 0.0);
		glEnd();
	}

	// Draw targets.
	vector<Sprite> targets = board.ConvertTargetsToSprites();

	for (unsigned int n = 0; n < targets.size(); n++) {
		Color targetColor = board.GetTargetColorAtIndex(targets[0].colorIndex);
		glColor3f(targetColor.r, targetColor.g, targetColor.b);
		float radiusX = targets[n].pixelWidth/2.0f;
		float radiusY = targets[n].pixelHeight/2.0f;

		glBegin(GL_POLYGON);
			for (double i = 0; i < 2 * M_PI; i += M_PI / CIRCLE_SEGMENTS)
				glVertex3f(targets[n].pixelX + cos(i) * radiusX, targets[n].pixelY + sin(i) * radiusY, 0.0);
		glEnd();
	}

	// Draw score.
	glColor3f(TEXT_COLOR.r, TEXT_COLOR.g, TEXT_COLOR.b);

	glPushMatrix();
	glTranslatef(0.025f, 0.935f, 0.0f);
	glScalef(0.0004f, 0.0004f, 0.0004f);

	char scoreChar[10];
	_itoa_s(score, scoreChar, 10, 10);
	string scoreString(scoreChar);

	for (unsigned int n = 0; n < scoreString.length(); n++)
		glutStrokeCharacter(GLUT_STROKE_ROMAN, scoreString[n]);

	glPopMatrix();

	glFlush();
}

// Draws game over message.
void DrawGameOver() {
	float scale;
	float lineWidth;
	int wordBeginIndex = 0;

	glClear(GL_COLOR_BUFFER_BIT);

	// Draw game over message.
	glColor3f(TEXT_COLOR.r, TEXT_COLOR.g, TEXT_COLOR.b);
	glPushMatrix();

	glTranslatef(0.0f, 0.7f, 0.0f);
	scale = 0.00075f;
	glScalef(scale, scale, scale);
	lineWidth = GetLineWidth(GAME_OVER_STRING, 0, GAME_OVER_STRING.length());
	glTranslatef((0.5f/scale) - (lineWidth/2.0f), 0.0f, 0.0f);
	glLineWidth(5.0f);

	for (unsigned int n = 0; n < GAME_OVER_STRING.length(); n++)
		glutStrokeCharacter(GLUT_STROKE_ROMAN, GAME_OVER_STRING[n]);

	glPopMatrix();

	// Draw final score.
	char scoreChar[10];
	_itoa_s(score, scoreChar, 10, 10);
	string scoreString = SCORE;
	scoreString += scoreChar;

	glPushMatrix();

	glTranslatef(0.0f, 0.5f, 0.0f);
	scale = 0.0004f;
	glScalef(scale, scale, scale);
	lineWidth = GetLineWidth(scoreString, 0, scoreString.length());
	glTranslatef((0.5f/scale) - (lineWidth/2.0f), 0.0f, 0.0f);
	glLineWidth(3.0f);

	for (unsigned int n = 0; n < scoreString.length(); n++)
		glutStrokeCharacter(GLUT_STROKE_ROMAN, scoreString[n]);

	glPopMatrix();

	// Draw restart instructions.
	glPushMatrix();

	glTranslatef(0.0f, 0.4f, 0.0f);
	scale = 0.00035f;
	glScalef(scale, scale, scale);
	glLineWidth(3.0f);

	for (unsigned int n = 0; n < RESTART.length();)
	{
		unsigned int lineEnd = RESTART.find('\n', n);

		if (lineEnd == -1)
			lineEnd = RESTART.length();
		
		lineWidth = GetLineWidth(RESTART, n, lineEnd);

		glTranslatef((0.5f/scale) - (lineWidth/2.0f), -0.075f/scale, 0.0f);

		for (unsigned int m = wordBeginIndex; m < lineEnd; m++)
			glutStrokeCharacter(GLUT_STROKE_ROMAN, RESTART[m]);

		glTranslatef(-lineWidth, 0.0f, 0.0f);
		glTranslatef(-((0.5f/scale) - (lineWidth/2.0f)), 0.0f, 0.0f);

		wordBeginIndex = lineEnd+1;
		n = lineEnd + 1;
	}

	glPopMatrix();

	glFlush();
}

// Doesn't do anything, but GLUT needs a display function.
void EmptyDisplayCallback() {

}

// Returns the length of the given string between startIndex and endIndex.
float GetLineWidth(string text, int startIndex, int endIndex) {
	float lineWidth = 0.0f;

	for (int n = startIndex; n < endIndex; n++)
		lineWidth += glutStrokeWidth(GLUT_STROKE_ROMAN, text[n]);

	return lineWidth;
}

// Handles keyboard input.
void CheckKeyboardInput(unsigned char key, int x, int y) {
	// Exit if escape key is pressed.
	if (key == 27)
	{
		glutDestroyWindow(windowID);
		exit(0);
		return;
	}

	if (currentState == MAIN_MENU)
	{
		// Change to game state if space key is pressed.
		if (key == ' ')
		{
			currentState = GAME;
			glutIdleFunc(GameMainLoop);
			glutSpecialFunc(CheckArrowInput);
		}
	}
	else if (currentState == GAME)
	{
		// Change direction of caterpillar.
		switch (key)
		{
		case 'w':
			caterpillar.ChangeDirection(UP);
			break;
		case 's':
			caterpillar.ChangeDirection(DOWN);
			break;
		case 'a':
			caterpillar.ChangeDirection(LEFT);
			break;
		case 'd':
			caterpillar.ChangeDirection(RIGHT);
			break;
		}
	}
	else if (currentState == GAME_OVER)
	{
		// Reset game if space key is pressed.
		if (key == ' ')
		{
			ResetGame();
		}
	}
}

// Handles arrow input.
void CheckArrowInput(int key, int x, int y) {
	// Change direction of caterpillar.
	switch (key)
	{
	case GLUT_KEY_UP:
		caterpillar.ChangeDirection(UP);
		break;
	case GLUT_KEY_DOWN:
		caterpillar.ChangeDirection(DOWN);
		break;
	case GLUT_KEY_LEFT:
		caterpillar.ChangeDirection(LEFT);
		break;
	case GLUT_KEY_RIGHT:
		caterpillar.ChangeDirection(RIGHT);
		break;
	}
}

// Entry point.
int main(int iArgc, char** cppArgv) {
	srand (time(NULL));

	// Start in the main menu state.
	currentState = MAIN_MENU;

	// Initialize window.
	InitializeView(PIXEL_WIDTH, PIXEL_HEIGHT);
	InitializeOpenGL();

	// Set up GLUT callbacks for main menu.
	glutIdleFunc(MainMenuLoop);
	glutDisplayFunc(EmptyDisplayCallback);
	glutKeyboardFunc(CheckKeyboardInput);
	glutMainLoop();
	return 0;
}

